﻿using System;
using SDX.Compiler;
using Mono.Cecil;
using Mono.Cecil.Cil;
using System.Linq;
using System.Collections.Generic;

public class PrefabSpawnMax : IPatcherMod
{
    public bool Link(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        InsertPregenerateHubCellDataCoroutine(gameModule, modModule);
        InsertLoadData(gameModule, modModule);
        InsertCW(gameModule, modModule);
        InsertOW(gameModule, modModule);
        InsertCopyPrefabData(gameModule, modModule);
        InsertAddRandomSpawnPointListHubCell(gameModule, modModule);
        InsertIncWorldPrefabCountWilderness(gameModule, modModule);
        InsertGetPrefabs(gameModule, modModule);
        InsertPrefabSpawnRuleGetPrefabs(gameModule, modModule);
        return true;
    }
    public void InsertIncWorldPrefabCountWilderness(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HubCell" && d.Namespace == "RWG2");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "YZ");

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = null;
        VariableDefinition vd = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void System.Collections.Generic.List`1<RWG2.HubCell/Lot>::Add(!0)")
            {
                if (inst.Previous.OpCode == OpCodes.Ldloc_S && inst.Previous.Operand.GetType() == typeof(VariableDefinition))
                {
                    vd = (VariableDefinition)inst.Previous.Operand;
                    insertPos = inst.Next;

                }
            }
        }

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition incWorldPrefabCount = prefabSpawnMaxMethod.Methods.First(d => d.Name == "IncWorldPrefabCount");
        MethodReference incWorldPrefabCountRef = gameModule.Import(incWorldPrefabCount);

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_S, vd));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, incWorldPrefabCountRef));

    }

    public void InsertAddRandomSpawnPointListHubCell(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "HGSockets" && d.Namespace == "RWG2.HubGenerators");


        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GenerateLots");


        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos1 = null;
        Instruction insertPos2 = null;
        VariableDefinition vd1 = null;
        VariableDefinition vd2 = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void System.Collections.Generic.List`1<RWG2.HubCell/Lot>::Add(!0)")
            {
                if (inst.Previous.OpCode == OpCodes.Ldloc_S && inst.Previous.Operand.GetType() == typeof(VariableDefinition))
                {
                    VariableDefinition vd = (VariableDefinition)inst.Previous.Operand;
                    if (vd.Index == 29)
                    {
                        insertPos1 = inst.Next;
                        vd1 = vd;
                    }
                    else
                    {
                        insertPos2 = inst.Next;
                        vd2 = vd;

                    }

                }
            }
        }

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition incWorldPrefabCount = prefabSpawnMaxMethod.Methods.First(d => d.Name == "IncWorldPrefabCount");
        MethodReference incWorldPrefabCountRef = gameModule.Import(incWorldPrefabCount);

        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Ldloc_S, vd1));
        proc1.InsertBefore(insertPos1, Instruction.Create(OpCodes.Call, incWorldPrefabCountRef));

        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Ldloc_S, vd2));
        proc1.InsertBefore(insertPos2, Instruction.Create(OpCodes.Call, incWorldPrefabCountRef));




    }


    public void InsertGetPrefabs(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "PrefabInfo" && d.Namespace == "RWG2.Rules");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GetPrefabs");

        Instruction insertPos = null;
        Instruction jumpToPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if(inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName== "System.Int32 RWG2.Rules.PrefabInfo::get_SpawnedCount()")
            {
                insertPos = inst.Previous;
                jumpToPos = (Instruction)inst.Next.Next.Next.Operand;
            }
        }

        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition isSpawn = prefabSpawnMaxMethod.Methods.First(d => d.Name == "IsSpawn");
        MethodReference isSpawnRef = gameModule.Import(isSpawn);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, isSpawnRef));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Brfalse, jumpToPos));

    }
    public void InsertPrefabSpawnRuleGetPrefabs(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "PrefabSpawnRule" && d.Namespace == "RWG2.Rules");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "GetPrefabs");

        Instruction insertPos = null;
        Instruction jumpToPos = null;
        foreach (Instruction inst in methodDef.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Int32 RWG2.Rules.PrefabInfo::get_SpawnedCount()")
            {
                insertPos = inst.Previous;
                jumpToPos = (Instruction)inst.Next.Next.Next.Next.Next.Operand;
            }
        }

        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition isSpawn = prefabSpawnMaxMethod.Methods.First(d => d.Name == "IsSpawn");
        MethodReference isSpawnRef = gameModule.Import(isSpawn);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldloc_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, isSpawnRef));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Brfalse, jumpToPos));

    }
    public void InsertCopyPrefabData(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == ".ctor" && d.Parameters.Count == 1);
        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = methodDef.Body.Instructions[0];

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition copyPrefabData = prefabSpawnMaxMethod.Methods.First(d => d.Name == "CopyPrefabData");
        MethodReference copyPrefabDataRef = gameModule.Import(copyPrefabData);
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_1));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, copyPrefabDataRef));

    }


    public void InsertCW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "CW");

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition getPrefabSpawnMaxProperty = prefabSpawnMaxMethod.Methods.First(d => d.Name == "GetPrefabSpawnMaxProperty");
        MethodReference getPrefabSpawnMaxPropertyRef = gameModule.Import(getPrefabSpawnMaxProperty);


        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction insertPos = methodDef.Body.Instructions[0];

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getPrefabSpawnMaxPropertyRef));

    }

    public void InsertOW(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition classDef = gameModule.Types.First(d => d.Name == "Prefab");
        MethodDefinition methodDef = classDef.Methods.First(d => d.Name == "OW");

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition getPrefabSpawnMaxToProperty = prefabSpawnMaxMethod.Methods.First(d => d.Name == "GetPrefabSpawnMaxToProperty");
        MethodReference getPrefabSpawnMaxToPropertyRef = gameModule.Import(getPrefabSpawnMaxToProperty);


        ILProcessor proc1 = methodDef.Body.GetILProcessor();

        Instruction insertPos = methodDef.Body.Instructions[0];

        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Ldarg_0));
        proc1.InsertBefore(insertPos, Instruction.Create(OpCodes.Call, getPrefabSpawnMaxToPropertyRef));

    }

    public void InsertPregenerateHubCellDataCoroutine(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition baseRWGDataLoader = gameModule.Types.First(d => d.Name == "BaseRWGDataLoader" && d.Namespace == "RWG2");
        TypeDefinition nestedClass = baseRWGDataLoader.NestedTypes.First(d => d.Name == "DH");
        MethodDefinition methodDef = nestedClass.Methods.First(d => d.Name == "MoveNext");

        ILProcessor proc1 = methodDef.Body.GetILProcessor();
        Instruction insertPos = null;
        bool isFirst = true;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Call && inst.Operand.GetType() == typeof(MethodReference) && ((MethodReference)inst.Operand).FullName == "System.Void Log::Out(System.String)")
            {
                if (isFirst)
                {
                    isFirst = false;
                }
                else
                {
                    insertPos = inst.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous.Previous;
                }

            }
        }

        List<Instruction> jumpInstList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction))
            {
                if (insertPos.Offset == ((Instruction)inst.Operand).Offset)
                {

                    jumpInstList.Add(inst);
                }
            }
        }

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition clearWorldCountPrefabDic = prefabSpawnMaxMethod.Methods.First(d => d.Name == "ClearWorldCountPrefabDic");
        MethodReference clearWorldCountPrefabDicRef = gameModule.Import(clearWorldCountPrefabDic);
        Instruction newJumpTo = Instruction.Create(OpCodes.Call, clearWorldCountPrefabDicRef);
        proc1.InsertBefore(insertPos, newJumpTo);

        foreach (Instruction inst in jumpInstList)
        {
            inst.Operand = newJumpTo;
        }

    }

    public void InsertLoadData(ModuleDefinition gameModule, ModuleDefinition modModule)
    {
        TypeDefinition world = gameModule.Types.First(d => d.Name == "World");
        MethodDefinition loadWorld = world.Methods.First(d => d.Name == "LoadWorld");

        ILProcessor proc1 = loadWorld.Body.GetILProcessor();

        Instruction insertPos = null;
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.OpCode == OpCodes.Callvirt && inst.Operand.GetType() == typeof(MethodDefinition) && ((MethodDefinition)inst.Operand).FullName == "System.Void WorldCreationData::Apply(World,WorldState)")
            {
                insertPos = inst.Previous.Previous.Previous.Previous.Previous;

            }
        }
        List<Instruction> jumpInstList = new List<Instruction>();
        foreach (Instruction inst in proc1.Body.Instructions)
        {
            if (inst.Operand != null && inst.Operand.GetType() == typeof(Instruction))
            {
                if (insertPos.Offset == ((Instruction)inst.Operand).Offset)
                {

                    jumpInstList.Add(inst);
                }
            }
        }

        TypeDefinition prefabSpawnMaxMethod = modModule.Types.First(d => d.Name == "PrefabSpawnMaxMethod");
        MethodDefinition initWorldCountPrefabDic = prefabSpawnMaxMethod.Methods.First(d => d.Name == "InitWorldCountPrefabDic");
        MethodReference initWorldCountPrefabDicRef = gameModule.Import(initWorldCountPrefabDic);

        Instruction newJumpTo = Instruction.Create(OpCodes.Call, initWorldCountPrefabDicRef);
        proc1.InsertBefore(insertPos, newJumpTo);
        foreach(Instruction inst in jumpInstList)
        {
            inst.Operand = newJumpTo;
        }
    }
    


    public void AddPrefabField(ModuleDefinition module)
    {
        TypeReference tr = module.Import(typeof(float));
        TypeDefinition prefab = module.Types.First(d => d.Name == "Prefab");

        tr = module.Import(typeof(int));
        AddFieldDefinition(prefab, "WorldSpawnMax", FieldAttributes.Public, tr);
    }
    public bool Patch(ModuleDefinition module)
    {
        Console.WriteLine("==PrefabSpawnMax Patcher===");
        AddPrefabField(module);
        return true;
    }

    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        classDef.Fields.Add(fd);

    }
    public void AddFieldDefinition(TypeDefinition classDef, string fieldName, FieldAttributes fieldAttribute, TypeReference fieldType, object constantvalue)
    {
        FieldDefinition fd = new FieldDefinition(fieldName, fieldAttribute, fieldType);
        fd.Constant = constantvalue;
        classDef.Fields.Add(fd);

    }
    private void SetFieldToPublic(FieldDefinition field)
    {
        field.IsFamily = false;
        field.IsPrivate = false;
        field.IsPublic = true;

    }

}
