﻿using System;
using System.Linq;
using System.Xml;
using System.Collections.Generic;
using UnityEngine;
using System.IO;

public static class PrefabSpawnMaxMethod
{
    public class WorldPrefabCount
    {
        public WorldPrefabCount()
        {

        }
        public WorldPrefabCount(Prefab prefab)
        {
            _maxCount = prefab.WorldSpawnMax;
            _nowCount = 0;
        }
        public void Inc()
        {
            _nowCount++;
        }

        public bool Check()
        {
            if (_maxCount > 0)
            {
                if (_nowCount < _maxCount)
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return true;
            }
        }
        private int _nowCount;
        public int NowCount
        {
            get
            {
                return _nowCount;
            }
            set
            {
                _nowCount = value;
            }
        }
        private int _maxCount;
        public int MaxCount
        {
            get
            {
                return _maxCount;
            }
            set
            {
                _maxCount = value;
            }
        }
    }

    private static Dictionary<string, WorldPrefabCount> _worldPrefabCountDic;
    public static void InitWorldCountPrefabDic()
    {
        _worldPrefabCountDic = new Dictionary<string, WorldPrefabCount>();
    }
    public static void ClearWorldCountPrefabDic()
    {
        _worldPrefabCountDic = null;
    }

    public static bool IsSpawn(RWG2.Rules.PrefabInfo prefabInfo)
    {
        if(prefabInfo == null)
        {
            return false;
        }
        if(_worldPrefabCountDic == null)
        {
            return false;
        }
        if(prefabInfo.FilteredList)
        {
            return true;
        }
        if(!_worldPrefabCountDic.ContainsKey(prefabInfo.Name))
        {
            Prefab prefab = RWG2.Rules.RWGRules.Instance.GetPrefab(prefabInfo.Name);
            if(prefab == null)
            {
                return false;
            }
            _worldPrefabCountDic.Add(prefabInfo.Name, new WorldPrefabCount(prefab));
        }
        return _worldPrefabCountDic[prefabInfo.Name].Check();

    }
    public static void IncWorldPrefabCount(RWG2.HubCell.Lot lot)
    {
        if (!_worldPrefabCountDic.ContainsKey(lot.PrefabName))
        {
            Prefab prefab = RWG2.Rules.RWGRules.Instance.GetPrefab(lot.PrefabName);
            _worldPrefabCountDic.Add(lot.PrefabName, new WorldPrefabCount(prefab));
        }
        _worldPrefabCountDic[lot.PrefabName].Inc();
    }

    //Prefab.CW
    public static void GetPrefabSpawnMaxProperty(Prefab prefab)
    {
        if (prefab.properties.Values.ContainsKey("WorldSpawnMax"))
        {

            prefab.WorldSpawnMax = int.Parse(prefab.properties.GetStringValue("WorldSpawnMax"));
        }
        else
        {
            prefab.WorldSpawnMax = 0;

        }

    }
    //Prefab.OW
    public static void GetPrefabSpawnMaxToProperty(Prefab prefab)
    {
        if (!(prefab.WorldSpawnMax == 0)
           )
        {
            prefab.properties.Values["WorldSpawnMax"] = prefab.WorldSpawnMax.ToString();
        }

    }
    public static void CopyPrefabData(Prefab src, Prefab dst)
    {
        dst.WorldSpawnMax = src.WorldSpawnMax;
    }

}
